import os
import sys
import logging
from ctypes import CDLL, c_bool, c_void_p

logger = logging.getLogger(__name__)
def module_path():
    if hasattr(sys, 'frozen'):
        fpath = sys.executable
    else:
        fpath = __file__
    return os.path.dirname(fpath)

try:
    arch = ''
    if sys.maxsize > 2**32:
        arch = '64'

    if sys.platform.startswith('win'):
        so = 'steam_api%s.dll' %arch
    elif sys.platform.startswith('linux'):
        so = 'libsteam_api%s.so' %arch
    elif sys.platform.startswith('darwin'):
        so = 'libsteam_api.dylib'
    else:
        logger.error('Unsupported operating system')
        raise EnvironmentError('Unsupported operating system')
    libpath = os.path.join(module_path(), 'libs', so)
    logger.debug('Looking for libsteam in "%s"', libpath)
    # Load the library
    api = CDLL(libpath)
    # Define return typed of "exported" functions
    api.SteamAPI_IsSteamRunning.restype = c_bool
    api.SteamAPI_Init.restype = c_bool
    api.SteamAPI_Shutdown.restype = c_void_p
except Exception as e:
    logger.exception('Loading steam library failed')
    print('Loading steam library failed: {}'.format(e))
    raise

SteamAPI_Init = api.SteamAPI_Init
IsSteamRunning = api.SteamAPI_IsSteamRunning
SteamAPI_Shutdown = api.SteamAPI_Shutdown



import os
import re
import stat
import shelve
import logging
import requests
import multiprocessing
from bs4 import BeautifulSoup
from steam_idle import GLOBAL_BLACKLIST

# Some regular expressions
re_AppId = re.compile(r'card_drop_info_gamebadge_(\d+)_')
re_Drops = re.compile(r'(\d+) card drop(?:s\b|\b) remaining')
re_PlayTime = re.compile(r'(\d+\.\d) hrs on record')

class PageParserError(Exception):
    pass
class AppIdBlacklistedError(PageParserError):
    pass
class AppIdNotFoundError(PageParserError):
    pass

class App(object):
    appid = None
    name = None
    remainingDrops = None
    playTime = None
    icon = property(lambda self: self._imgpath('icon'))
    logosmall = property(lambda self: self._imgpath('logosmall'))
    header = property(lambda self: self._imgpath('header'))
    def __init__(self, image_path=''):
        self.image_path = image_path
    def __repr__(self):
        return '<[{:6d}] "{}" ({}, {:.1f})>'.format(
            self.appid or 0,
            self.name.encode('ascii', 'ignore') or 'Unknown app',
            -1 if self.remainingDrops == None else self.remainingDrops,
            -1.0 if self.playTime == None else self.playTime,
        )
    def __eq__(self, other):
        return isinstance(other, self.__class__) and \
            self.appid == other.appid and \
            self.name == other.name and \
            self.remainingDrops == self.remainingDrops and \
            self.playTime == self.playTime
    def __hash__(self):
        return hash((
            self.appid,
            self.name,
            self.remainingDrops,
            self.playTime
        ))
    def _imgpath(self, imgtype):
        if not isinstance(self.appid, int):
            return None
        return os.path.join(self.image_path, '%d_%s.jpg' % (self.appid, imgtype))
    @property
    def storeUrl(self):
        return 'https://store.steampowered.com/app/{}'.format(self.appid or '')

def mockSome():
    import random
    apps = {}
    for i in (232770,262830,285010,321950,343100,307170,311210,233290):
        a = App()
        a.appid = i
        a.remainingDrops = random.randint(0,6)
        a.playTime = round(random.uniform(0.0, 10.0),1)
        apps[i] = a
    return apps

class FetchImages(multiprocessing.Process):
    ''' Multiprocessing worker to fetch and store icon and logo for an app
    '''
    def __init__(self, task_queue, image_path):
        super(FetchImages, self).__init__()
        self.logger = logging.getLogger('.'.join((__name__, self.__class__.__name__)))
        self.task_queue = task_queue
        self.image_path = image_path
        self.session = requests.Session()

    def run(self):
        while True:
            next_task = self.task_queue.get()
            if next_task is None:
                # Poison pill
                self.task_queue.task_done()
                self.session.close()
                break
            # Process this task
            self.fetch_images(next_task)
            self.task_queue.task_done()
        self.session.close()
        return

    def fetch_images(self, task):
        appinfo = task
        appid = appinfo.get('appid')
        for imgtype in ('icon', 'logosmall', 'header'):
            filename = '%d_%s.jpg' % (appid, imgtype)
            imagepath = os.path.join(self.image_path, filename)
            if imgtype == 'header':
                url = 'https://steamcdn-a.akamaihd.net/steam/apps/%d/header_292x136.jpg' % appid
            else:
                url = appinfo.get(imgtype+'url')
            self.logger.debug('Processing %d: imgtype: %s, filename: "%s", path: "%s", url: "%s"',
                                appid, imgtype, filename, imagepath, url)
            if url is None:
                self.logger.error('No URL found')
                continue

            r = self.session.get(url)
            with open(imagepath, 'wb') as f:
                f.write(r.content)

def chunks(l, n):
    '''Yield successive n-sized chunks from l.'''
    for i in range(0, len(l), n):
        yield l[i:i+n]

class SteamBadges(object):
    ''' Holds methods for parsing badge pages etc. '''
    def __init__(self, swb, data_path=''):
        self.logger = logging.getLogger('.'.join((__name__, self.__class__.__name__)))
        self.swb = swb
        if data_path != '':
            self.data_path = data_path
        else:
            self.data_path = swb.appdata_path
        self.logger.info('Using "%s" as data path', self.data_path)
        self.shelve_path = os.path.join(self.data_path, 'cache')
        self.image_path = os.path.join(self.data_path, 'images')
        if not os.path.exists(self.image_path):
            os.makedirs(self.image_path, stat.S_IRWXU)

    def parse_badge(self, badge):
        app = App(self.image_path)
        try:
            # Parse AppId
            drop_info = badge.find('div', {'class': 'card_drop_info_dialog'}).attrs.get('id')
            app.appid = int(re_AppId.match(drop_info).groups()[0])
        except:
            raise AppIdNotFoundError('Could not parse AppId from badge: %s' % badge.prettify())

        if app.appid in GLOBAL_BLACKLIST: #TODO Blacklist check should be done by the "frontend"
            raise AppIdBlacklistedError('%d is in global blacklist' % app.appid)

        try:
            # Parse remaining drops (will raise if there are none)
            progress = badge.find('span', {'class': 'progress_info_bold'}).get_text()
            app.remainingDrops = int(re_Drops.match(progress).groups()[0])
        except:
            app.remainingDrops = 0

        try:
            # Parse play time
            app.playTime = float(re_PlayTime.search(badge.get_text()).groups()[0])
        except:
            app.playTime = 0.0

        return app

    def parse_badges_pages(self, appid_filter=None):
        ''' Iterates over all badges pages of a steam profile
            Parses all badges (using parse_badge()) to return the appId, play time ('till now) and the number of card drops left.

            @param appid_filter only look for appids listed here
        '''
        appid_filter = appid_filter or []
        filter_appids = True if appid_filter else False
        parsed_apps = {}
        currentPage = badgePages = 1

        retry = False
        while currentPage <= badgePages and (filter_appids == False or len(appid_filter) > 0):
            r = self.swb.get('https://steamcommunity.com/my/badges', params={'p': currentPage})
            if r.status_code == 302:
                if retry:
                    # We already tries to force a login
                    raise Exception('Unable to fetch badges')
                # Looks like we've been redirected. Force a login and retry
                print('Need to login again')
                self.swb.login()
                retry = True
                continue

            soup = BeautifulSoup(r.content, 'html.parser')
            if currentPage == 1:
                try:
                    badgePages = int(soup.find_all('a', {'class': 'pagelink'})[-1].get_text())
                except:
                    pass

            for b in soup.find_all('div', {'class': 'badge_title_stats'}):
                try:
                    app = self.parse_badge(b)
                except PageParserError:
                    # Could not correctly parse app info, continue with the next one
                    continue

                # AppId's where given as filter, check if this AppId is one of those
                if filter_appids:
                    if app.appid in appid_filter:
                        appid_filter.remove(app.appid)
                    else:
                        # This appid is NOT in the filter list.
                        # don't include it in the returned list
                        continue

                # Add app info to the list of parsed badges
                parsed_apps[app.appid] = app

                if filter_appids and len(appid_filter) == 0:
                    # AppId's where given as filter and all of them where found already
                    # so we are done.
                    break
            # Continue with next page
            currentPage += 1

        if not parsed_apps:
            #FIXME: empty badges page == no good.
            print('ERRPOR: Could not find any badges on badge page')
            from tempfile import NamedTemporaryFile
            tmpf = NamedTemporaryFile(prefix='quickDump_', suffix='.html', delete=False)
            print('Dumping r.content to: "file://%s"' % tmpf.name)
            tmpf.write(r.content)
            tmpf.close()
            rname = tmpf.name.replace('.html', '.dat')
            with open(rname, 'w') as tmpf:
                print('Dumping r to: "file://%s"' % rname)
                tmpf.write(vars(r))

        return parsed_apps

    def drop_app_cache(self, appid):
        ''' Drop all info about an app (images from filesystem and app info from shelve)
        '''
        with shelve.open(self.shelve_path) as appshelve:
            try:
                del appshelve[str(appid)]
            except KeyError:
                # No such key in shelve, ignore
                pass
        for imgtype in ('icon', 'logosmall'):
            filename = '%d_%s.jpg' % (appid, imgtype)
            imagepath = os.path.join(self.image_path, filename)
            if os.path.exists(imagepath):
                os.unlink(imagepath)

    def get_apps(self, appid_filter=None, fetch_images=True):
        ''' Parse the badge pages, add app info (like name and icon) if needed
            fetch and store the icons and cache app info in shelve.

            Return a dict of all apps on badges page (with and without remaining drops):
            {<appid>: <App istance>, <appid>: <App instance>, ...}
        '''
        appid_filter = appid_filter or []
        apps = self.parse_badges_pages(appid_filter)
        #apps = mockSome()

        # check for appids not in shelve
        appshelve = shelve.open(self.shelve_path)
        appids_not_in_shelve = []
        for appid in apps.keys():
            if str(appid) in appshelve:
                # Load aditional app info from shelve
                a = appshelve[str(appid)]
                apps[appid].name = a['name']
            else:
                appids_not_in_shelve.append(str(appid))

        if appids_not_in_shelve:
            # GetAppInfo only returns info for 100 apps at once
            # TODO: Call https://api.steampowered.com/IPlayerService/GetOwnedGames/v0001/ with access_token instead?
            # params = {'access_token': self.swb.oauth_access_token, 'appids_filter': ','.join(appids_not_in_shelve), 'steamid': swb.steamid, 'format': 'json', 'include_appinfo': 1}

            appinfos = []
            self.logger.debug('Requesting %d appids from GetAppInfo:', len(appids_not_in_shelve))
            for appid_chunk in chunks(appids_not_in_shelve, 100):
                params = {
                    'access_token': self.swb.oauth_access_token,
                    'appids': ','.join(appid_chunk)
                }
                self.logger.debug('Requesting a chunk of %d appids from GetAppInfo:', len(appid_chunk))
                r = self.swb.get('https://api.steampowered.com/ISteamGameOAuth/GetAppInfo/v1/', params=params)
                ainfo = r.json().get('apps', [])
                appinfos.extend(ainfo)
                self.logger.debug('GetAppInfo returned data for %d appids:', len(ainfo))

            if fetch_images is True:
                # Retrieve and store icon and logosmall
                tasks = multiprocessing.JoinableQueue()
                num_consumers = multiprocessing.cpu_count() * 2

                consumers = []
                for _i in range(num_consumers):
                    c = FetchImages(tasks, self.image_path)
                    c.start()
                    consumers.append(c)

                for appinfo in appinfos:
                    tasks.put(appinfo)

                # Add a poison pill for each consumer
                for _i in range(num_consumers):
                    tasks.put(None)

                # Wait for all of the tasks to finish
                tasks.join()

            # Merge new data with data from shelve and store new values
            for appinfo in appinfos:
                appid = appinfo.get('appid')
                name = appinfo.get('name')
                apps[appid].name = name
                # Store in shelve
                data = {
                    'name': name,
                }
                appshelve[str(appid)] = data

        appshelve.close()
        return apps



#!/usr/bin/env python
from __future__ import print_function
import os
import sys
import argparse
import atexit
from time import sleep
from datetime import timedelta, datetime
from math import ceil
import logging
from steamweb import SteamWebBrowserCfg as SteamWebBrowser
from steam_idle import steam_api
from steam_idle.page_parser import SteamBadges
from steam_idle.idle import strfsec, IdleChild, calc_delay

def main_idle(apps):
    # Just to make sure it's ordered
    apps = sorted(apps, key=lambda x: x.playTime, reverse=True)
    if not args.skip_multi:
        # Idle all apps with playTime < 2h in parallel
        processes = []
        for app in [app for app in apps if app.playTime < 2.0]:
            delay = int((2.0 - app.playTime) * 60 * 60)
            endtime = (datetime.now() + timedelta(seconds=delay))
            p = IdleChild(app)
            p.start()
            processes.append((endtime, p))
            if args.verbose:
                print(p.name)
            # Steam client will crash if childs spawn too fast; Fixes #1
            sleep(0.25)
        if args.verbose:
            print('Multi-Idling %d apps' % len(processes))

        # Should be ordered, shortest idle first
        for endtime, p in processes:
            now = datetime.now()
            if endtime < now:
                print(p, 'endtime (%s) is in the past, shutting down' % (endtime,))
                p.terminate()
                p.join()
                continue
            diff = int(ceil((endtime - now).total_seconds()))
            if diff <= 0:
                print(p, 'diff (%s) is below 0, shutting down' % (diff,))
                p.terminate()
                p.join()
                continue
            print('Sleeping for %s till %s' %(
                strfsec(diff),
                (datetime.now() + timedelta(seconds=diff)).strftime('%c')
            ))
            sleep(diff)
            if args.verbose:
                print(p, 'Woke up, shutting down')
            p.terminate()
            p.join()

        if processes:
            # Multi-Idled some apps, update values as they will have changed
            apps = [app for app in sbb.get_apps(fetch_images=False).values() if app.remainingDrops > 0]
            # If there are still apps with < 2.0h play time, restart
            if [app for app in apps if app.playTime < 2.0] > 1:
                print('There are still apps within refund time, restarting multi-idle')
                return main_idle(apps)

    # All apps should be out of refund time, (playTime >= 2h), idle one by one
    if args.verbose:
        print('Startin sequential idle of %d apps' % len(apps))
    new_apps = [] # new apps added douring idle
    for app in apps:
        p = IdleChild(app)
        p.start()
        while app.remainingDrops > 0:
            delay = calc_delay(app.remainingDrops)
            print('"%s" has %d remaining drops: Idling for %s (\'till %s)' % (
                    app.name,
                    app.remainingDrops,
                    strfsec(delay),
                    (datetime.now() + timedelta(seconds=delay)).strftime('%c')
            ))
            sleep(delay)

            # Re check for remainingDrops and new apps
            for a in sbb.get_apps(fetch_images=False).values():
                if a.remainingDrops > 0 and not [x for x in apps + new_apps if x.appid == a.appid]:
                    print('Found a new app to idle: "%s" has %d remaining drops, play time till now: %0.1f hours' % (a.name, a.remainingDrops, a.playTime))
                    if a.playTime >= 2.0:
                        # Already out of refund time, add to the one by one idle list
                        apps.append(a)
                    else:
                        # Needs "refund-idle"
                        new_apps.append(a)
                    continue

                if a.appid == app.appid:
                    # Still drops left for this app, continue idleing
                    app = a
            print('%d drops remaining, playtime is %0.1f' %(app.remainingDrops, app.playTime))

        # Stop idleing
        p.terminate()
        p.join()

    if new_apps:
        # We've found new apps that need "refund-idle"
        print('Found %d new apps that need "refund-idle":' %(len(new_apps),))
        if args.verbose:
            for app in new_apps:
                print('"%s" has %d remaining drops, play time till now: %0.1f hours' % (app.name, app.remainingDrops, app.playTime))
        return main_idle(new_apps)

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Idle all steam apps with card drops left.')
    parser.add_argument('-v', '--verbose', help='increase output verbosity', action='store_true')
    parser.add_argument('-d', '--debug', help='enable debug output', action='store_true')
    parser.add_argument('-l', '--list', help='don\'t idle, just list apps with card drops', action='store_true')
    parser.add_argument('--skip-multi', help='don\'t multi-idle all apps with playtime < 2h first', action='store_true')
    parser.add_argument('-a', '--appid', help='idle only specific app ID\'s', type=int, nargs='*')
    args = parser.parse_args()

    if args.debug:
        logging.basicConfig(format='%(asctime)s (%(name)s.%(funcName)s) [%(levelname)s] %(message)s', level=logging.DEBUG)

    swb = SteamWebBrowser()
    sbb = SteamBadges(swb)

    if not args.list:
        # make sure this is only run once
        pidfile = os.path.join(os.path.dirname(__file__), 'steam-idle.pid')
        if os.path.isfile(pidfile):
            print('already running ("%s")' % pidfile)
            sys.exit(1)
        with open(pidfile, 'w') as pf:
            pf.write(str(os.getpid()))
        atexit.register(os.unlink, pidfile)

        # Check if Steam is running
        if not steam_api.IsSteamRunning():
            print('Could not find a running Steam instance!')
            print('Please start your Steam Client.')
            sys.exit(1)

    apps = [app for app in sbb.get_apps(fetch_images=False).values() if app.remainingDrops > 0]

    print('%d games with a total of %d card drops left:' % (len(apps), sum([app.remainingDrops for app in apps])))
    if args.verbose or args.list:
        for app in apps:
            print('"%s" has %d remaining drops, play time till now: %0.1f hours' % (app.name, app.remainingDrops, app.playTime))
    if args.list:
        sys.exit(0)

    main_idle(apps)



import sys
import os
import multiprocessing
import setproctitle
from datetime import timedelta
from time import sleep
from steam_idle import steam_api

strfsec = lambda seconds: str(timedelta(seconds=seconds))
def r_sleep(sec):
    ''' Sleep sec seconds and return seconds slept '''
    sleep(sec)
    return sec

class IdleChild(multiprocessing.Process):
    def __init__(self, app):
        super(IdleChild, self).__init__()
        self.app = app
        self.name += '-[%s]' % self.app.name if self.app.name else str(self.app.appid)

    def run(self):
        setproctitle.setproctitle(self.name)
        os.environ['SteamAppId'] = str(self.app.appid)
        self.redirect_streams()
        try:
            steam_api.SteamAPI_Init()
            self.restore_streams()
        except:
            self.restore_streams()
            p = multiprocessing.current_process()
            print('%s(%d): Couldn\'t initialize Steam API' % (p.name, p.pid))
            sys.stdout.flush()
            return

        while True:
            sleep(1)

    def redirect_streams(self):
        # redirect stdout and stderr of steam api
        devnull = os.open(os.devnull, 777)
        self.old_stdout = os.dup(1)
        self.old_stderr = os.dup(2)
        os.dup2(devnull, 1)
        os.dup2(devnull, 2)

    def restore_streams(self):
        # restore stdout and stderr
        os.dup2(self.old_stdout, 1)
        os.dup2(self.old_stderr, 2)

sameDelay = 0
lastDelay = 5
def calc_delay(remainingDrops):
    ''' Calculate the idle delay
        Minimum play time for cards to drop is ~20min again. Except for accounts
        that requested a refund?

        Re-check every 15 mintes if there are more than 1 card drops remaining.
        If only one drop remains, check every 5 minutes
    '''
    global sameDelay, lastDelay

    # Reset lastDelay for new appids
    if remainingDrops > 1:
        lastDelay = 5
        sameDelay = 0

    if remainingDrops > 2:
        return 15 * 60 # Check every 15 minutes
    elif remainingDrops == 2:
        return 10 * 60 # Check every 10 minutes
    else:
        # decrease delay by one minute every two calls
        if lastDelay > 1:
            if sameDelay == 2:
                sameDelay = 0
                lastDelay -= 1
            sameDelay += 1
        return lastDelay * 60 # Check every 5 minutes



#!/usr/bin/env python

from setuptools import setup

def parse_requirements(path):
    with open(path, 'r') as infile:
        return [l.strip() for l in infile.readlines()]

setup(
    name = 'steam_idle',
    version = '0.2',
    description = 'Idle Steam apps/games for card drops',
    long_description = open('README.rst', 'r').read(),
    platforms = ['any'],
    keywords = 'steam',
    author = 'Jayme',
    author_email = 'tuxnet@gmail.com',
    url = 'https://github.com/jayme-github/steam_idle',
    license = 'GNU Affero General Public License v3',
    classifiers = [
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: GNU Affero General Public License v3',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
    ],
    install_requires = parse_requirements('requirements.txt'),
    packages = ['steam_idle'],
    package_dir = {'steam_idle': 'steam_idle'},
    package_data = {'steam_idle':
                    ['libs/steam_api64.dll',
                    'libs/libsteam_api.so',
                    'libs/steam_api.dll',
                    'libs/libsteam_api64.so',
                    'libs/libsteam_api.dylib']
    },
    scripts = ['steam_idle_cli.py'],
)
